# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

import re
from typing import Tuple, Any, Optional, Union
from spx_sdk.components import SpxComponent
from spx_sdk.attributes import SpxAttribute
from spx_sdk.attributes.attribute import InternalAttributeWrapper, ExternalAttributeWrapper, StaticAttributeWrapper


# Define a regex pattern to match attribute references
ATTRIBUTE_REFERENCE_PATTERN = re.compile(r'[#\$@](attr|internal|external|in|out|ext)\(([^)]+)\)')


def is_attribute_reference(reference: str) -> Tuple[bool, str, Optional[str]]:
    """
    Check if the reference is an attribute reference and parse it.

    Args:
        reference (str): The reference string to check.

    Returns:
        Tuple[bool, str, Optional[str]]:
            - bool: whether `reference` is an attribute reference.
            - str: the parsed attribute string.
            - Optional[str]: the reference type ('attr', 'internal', or 'external'), or None.
    """
    is_attr = False
    attr_type = None
    attr_string = reference

    if isinstance(reference, str):
        match = ATTRIBUTE_REFERENCE_PATTERN.match(reference)
        if match:
            is_attr = True
            raw_type = match.group(1)
            attr_string = match.group(2)
            # Map to 'external' only if raw_type indicates external
            if raw_type in ("external", "out", "ext"):
                attr_type = "external"
            else:
                attr_type = "internal"
        elif '.' in reference:
            is_attr = True
            attr_type = "internal"
            attr_string = reference

    return is_attr, attr_string, attr_type


def find_attribute(instance: SpxComponent, attribute_string: str) -> Optional[SpxAttribute]:
    """
    Find and return the attribute object based on a simple name or a dot-separated path.

    Delegates:
      - dot-containing strings → nested lookup across children/instances
      - simple names → lookup in the 'attributes' container of the instance
    """
    if '.' in attribute_string:
        return _find_nested_attribute(instance, attribute_string)
    else:
        return _find_simple_attribute(instance, attribute_string)


def _find_simple_attribute(instance: SpxComponent, attr_name: str) -> Optional[SpxAttribute]:
    """
    Lookup an attribute by name in the 'attributes' container of the given instance.
    """
    attrs_cont = instance.get("attributes", None)
    if attrs_cont is None:
        raise ValueError(f"Component '{instance.name}' has no 'attributes' container.")
    return attrs_cont.get(attr_name, None)


def _find_nested_attribute(instance: SpxComponent, attribute_string: str) -> Optional[SpxAttribute]:
    """
    Lookup an attribute via a dot-separated path through component children, falling back
    to an 'instances' container if necessary, and finally in the 'attributes' container.
    """
    parts = attribute_string.split('.')
    current = instance

    # Traverse through component children for all but the final segment
    for segment in parts[:-1]:
        next_comp = current.get(segment, None)
        if next_comp is None:
            instances_cont = current.get("instances", None)
            if instances_cont and segment in instances_cont:
                next_comp = instances_cont.get(segment)
            else:
                return None
        current = next_comp

    # Final segment should be an attribute name
    attr_name = parts[-1]
    attrs_cont = current.get("attributes", None)
    if attrs_cont is None:
        raise ValueError(f"Component '{current.name}' has no 'attributes' container.")
    return attrs_cont.get(attr_name, None)


def resolve_attribute_reference(
    instance: SpxComponent,
    attribute_name: str
) -> Optional[Union[InternalAttributeWrapper, ExternalAttributeWrapper]]:
    """
    Resolve the attribute reference and return the attribute object and its type.

    Args:
        instance: The instance containing the attribute.
        attribute_name (str): The attribute reference string.

    Returns:
        Optional[Union[InternalAttributeWrapper, ExternalAttributeWrapper]]:
            A wrapper for the resolved attribute, or None if not an attribute reference.
    """
    is_attr, attr_string, attr_type = is_attribute_reference(attribute_name)
    if not is_attr:
        return None
    attribute = find_attribute(instance, attr_string)
    if attribute is None:
        raise ValueError(f"Attribute '{attr_string}' not found in instance '{instance.name}'.")
    return attribute.get_wrapper(attr_type)


def get_attribute_value(attribute: SpxAttribute, attr_type: str) -> Any:
    """
    Get the value of the attribute based on its type.

    Args:
        attribute (Attribute): The attribute object.
        attr_type (str): The attribute type ('internal' or 'external').

    Returns:
        Any: The attribute value.

    Raises:
        ValueError: If the attribute is None.
    """
    if attribute is None:
        raise ValueError("Attribute cannot be None when getting its value.")
    if attr_type == "internal":
        return attribute.internal_value
    else:
        return attribute.external_value


def set_attribute_value(attribute: SpxAttribute, attr_type: str, value: Any) -> None:
    """
    Set the value of the attribute based on its type.

    Args:
        attribute (Attribute): The attribute object.
        attr_type (str): The attribute type ('internal' or 'external').
        value (Any): The value to set.

    Raises:
        ValueError: If the attribute is None.
    """
    if attribute is None:
        raise ValueError("Attribute cannot be None when setting its value.")
    if attr_type == "internal":
        attribute.internal_value = value
    else:
        attribute.external_value = value


def resolve_reference(instance: SpxComponent, reference: str) -> Any:
    """
    Resolve the reference to its actual value.

    Args:
        instance: The instance containing the reference.
        reference (str): The reference string.

    Returns:
        Any: The resolved value if it was an attribute reference, otherwise the original `reference` string.
    """
    is_attr, attr_string, attr_type = is_attribute_reference(reference)
    if is_attr:
        attribute = find_attribute(instance, attr_string)
        if attribute is None:
            return reference
        return get_attribute_value(attribute, attr_type)
    return reference


def resolve_attribute_reference_hierarchical(
    instance: SpxComponent,
    reference: str
) -> Optional[Union[InternalAttributeWrapper, ExternalAttributeWrapper]]:
    """
    Resolve an attribute reference by searching up the component hierarchy.
    Starting from `instance`, traverse parent links until an attributes container
    with the named attribute is found. Returns the corresponding wrapper or None.
    """
    # Parse reference syntax
    is_attr, attr_string, attr_type = is_attribute_reference(reference)
    if not is_attr:
        return None

    # Walk up from the given instance to root
    current = instance
    while current is not None:
        # Attempt to find attribute on this component
        try:
            attribute = find_attribute(current, attr_string)
        except ValueError:
            attribute = None

        if attribute is not None:
            return attribute.get_wrapper(attr_type)
        current = current.parent  # move to next parent

    return None


def substitute_attribute_references_hierarchical(
    instance: SpxComponent,
    text: str
) -> str:
    """
    Replace all attribute reference patterns in `text` with their resolved values.
    Uses hierarchical resolution on `$attr(...)`, `$ext(...)`, etc., starting from `instance`.
    Args:
        instance (SpxComponent): Component context to resolve attribute references.
        text (str): The input string containing reference patterns.

    Returns:
        str: A new string with each reference replaced by repr(value).
    """
    def _repl(match: re.Match) -> str:
        ref = match.group(0)
        wrapper = resolve_attribute_reference_hierarchical(instance, ref)
        if wrapper is None:
            # Leave unknown refs unchanged
            return ref
        try:
            value = wrapper.get()
        except Exception:
            # Fallback: return original ref on error
            return ref
        # Use repr to preserve data types in embedded expressions
        return repr(value)

    return ATTRIBUTE_REFERENCE_PATTERN.sub(_repl, text)


def extract_attribute_wrappers_hierarchical(
    instance: SpxComponent,
    text: str
) -> list[tuple[str, Optional[Union[InternalAttributeWrapper, ExternalAttributeWrapper, StaticAttributeWrapper]]]]:
    """
    For each attribute reference or nested chain in `text`, resolve its wrapper hierarchically.
    Returns a list of (ref, wrapper) tuples, in order of appearance.
    """
    results = []
    # Patterns to match simple attrs and nested chains
    nested_pattern = re.compile(r'\$\(\.[^)]+\)')
    simple_pattern = ATTRIBUTE_REFERENCE_PATTERN

    # Gather all matches with their positions and type
    matches = []
    for m in simple_pattern.finditer(text):
        matches.append((m.start(), m.group(0), 'simple'))
    for m in nested_pattern.finditer(text):
        matches.append((m.start(), m.group(0), 'nested'))

    # Sort by appearance in text
    matches.sort(key=lambda x: x[0])

    for _, ref, kind in matches:
        if kind == 'nested':
            try:
                wrapper = resolve_nested_chain_reference(instance, ref)
            except Exception:
                wrapper = None
        else:
            try:
                wrapper = resolve_attribute_reference_hierarchical(instance, ref)
            except Exception:
                wrapper = None
        results.append((ref, wrapper))
    return results


def substitute_with_wrappers(
    text: str,
    wrappers: list[tuple[str, Optional[Union[InternalAttributeWrapper, ExternalAttributeWrapper]]]]
) -> str:
    """
    Replace all occurrences of each reference in `text` with repr(wrapper.get()) if possible.
    Leaves refs unchanged if wrapper is None or .get() raises.
    """
    result = text
    for ref, wrapper in wrappers:
        if wrapper is not None:
            try:
                value = wrapper.get()
            except Exception:
                continue
            # Replace all exact occurrences of ref with repr(value)
            result = result.replace(ref, repr(value))
    return result


# --- Nested chain reference resolver ---
def resolve_nested_chain_reference(
    instance: SpxComponent,
    reference: str
) -> Any:
    """
    Resolve a nested chain reference of the form '$(.comp1.comp2.attr)'.
    Starts from the root component and follows each dot-separated segment.
    Returns a StaticAttributeWrapper for the final attribute/property.
    """
    # Check pattern: must start with '$(' and end with ')'
    if not (isinstance(reference, str) and reference.startswith('$(') and reference.endswith(')')):
        raise ValueError(f"Invalid nested reference syntax: {reference!r}")
    # Extract inner path and ensure it begins with '.'
    path = reference[2:-1]
    if not path.startswith('.'):
        raise ValueError(f"Nested reference must start with '.': {path!r}")
    segments = path.lstrip('.').split('.')
    # Traverse all but the last segment to reach the target container
    current: Any = instance.get_root()
    for seg in segments[:-1]:
        if hasattr(current, 'children') and seg in current.children:
            current = current.children[seg]
        elif hasattr(current, seg):
            current = getattr(current, seg)
        else:
            raise ValueError(f"Cannot resolve segment '{seg}' in nested reference '{reference}'")
    # The last segment refers to an attribute or property on 'current'
    last_seg = segments[-1]
    if not hasattr(current, last_seg):
        raise ValueError(f"Cannot resolve final segment '{last_seg}' in nested reference '{reference}'")
    # Return a StaticAttributeWrapper for this attribute/property
    return StaticAttributeWrapper(current, last_seg)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from .conditions import Conditions, IfChain, Condition, Else

__all__ = ["Conditions", "IfChain", "Condition", "Else"]



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from spx_sdk.registry import register_class
from spx_sdk.components import SpxComponent


@register_class(name="refresh_model")
class RefreshHook(SpxComponent):
    def run(self, *args, **kwargs):
        self.get_root().prepare(*args, **kwargs)
        return self.get_root().run(*args, **kwargs)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from .actions import Action,  Actions
from .function_action import FunctionAction
from .set_action import SetAction

__all__ = [
    "Action",
    "Actions",
    "FunctionAction",
    "SetAction",
]



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

import time
import threading
import logging
from typing import Any, Dict

from spx_sdk.registry import register_class
from spx_sdk.components import SpxComponent


@register_class(name="simple_polling")
class SimplePolling(SpxComponent):
    """
    SimplePolling component that periodically invokes its parent component's run() method
    at a fixed interval.
    Configuration key:
      - interval (float): sleep interval in seconds between polls (default 0.1)
    """
    def _populate(self, definition: Dict[str, Any]) -> None:
        """
        Initialize interval and state from definition.
        """
        # Store definition
        self.definition = definition or {}
        # Initialize polling parameters
        self.reset()

    def reset(self, *args, **kwargs) -> bool:
        """
        Initialize or reinitialize polling parameters.
        """
        self.interval = self.definition.get("interval", 0.1)
        self.running = False
        self.thread = None
        return True

    def start(self, *args, **kwargs) -> bool:
        if not self._enabled:
            self.logger.debug(f"Component {self.name} is disabled; skipping start")
            return False
        if self.running:
            logging.warning("Polling is already running.")
            return False
        self.running = True
        self.thread = threading.Thread(target=self._run, daemon=True)
        self.thread.start()
        return True

    def run(self, *args, **kwargs):
        """
        Hook into parent.run() loop; does nothing by default.
        """
        return True

    def stop(self, *args, **kwargs) -> bool:
        """
        Stop the polling loop and wait for thread to finish.
        """
        self.running = False
        if self.thread is not None:
            self.thread.join()
            self.thread = None
        return True

    def _run(self):
        """
        Internal polling loop: call parent.run() then sleep for interval.
        """
        try:
            while self.running:
                if self.parent:
                    self.parent.run()
                time.sleep(self.interval)
        except Exception as e:
            logging.error(f"SimplePolling encountered error: {e}")
        finally:
            self.running = False



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from typing import Any, Optional, Type

from spx_sdk.registry import class_registry, create_instance, get_classes_by_base, register_class
from spx_sdk.components import SpxComponent


@register_class()
class SpxContainer(SpxComponent):
    """
    SpxContainer extends SpxComponent by automatically instantiating and organizing
    child components based on a provided definition.

    Two modes of operation:

    1. Filtered mode (when `type` is provided):
       - The definition may be a dict, list, or scalar.
       - For dict/list entries, if the key matches a subclass of `type`, that subclass
         is used; otherwise the base `type` class is used.
       - Scalar definitions are instantiated directly as the base `type`.

    2. Generic mode (when no `type` is provided):
       - Treat each dict key or single-key list node as a class name looked up in the registry.
       - Instantiate each matching class.
       - Other values (e.g. scalars) are ignored in generic mode.

    Child components are created during initialization by passing `definition` to `_populate`.
    """

    def __init__(
        self,
        definition: Any,
        *,
        name: Optional[str] = None,
        parent: Optional[SpxComponent] = None,
        type: Optional[Type[SpxComponent]] = None
    ):
        self._type = type
        # now actually load children
        super().__init__(name=name, parent=parent, definition=definition)

    def _populate(self, definition: Any) -> None:
        if self._type:
            self._load_filtered(definition)
        else:
            self._load_generic(definition)
        super()._populate(definition)

    def _handle_dict_filtered(self, definition: dict, base_cls, subclasses):
        """Instantiate children for dict-based filtered definitions."""
        # Iterate over a static list of keys to avoid modifying while iterating
        for cls_name in list(definition.keys()):
            cfg = definition.pop(cls_name)
            cls = subclasses.get(cls_name, base_cls)
            cls(name=cls_name, parent=self, definition=cfg)

    def _handle_list_filtered(self, definition: list, base_cls, subclasses):
        """Instantiate children for list-based filtered definitions, ensuring unique names."""
        # Track how many times each class name has been seen
        name_counts: dict[str, int] = {}
        original_len = len(definition)
        for _ in range(original_len):
            node = definition.pop(0)
            if isinstance(node, dict) and len(node) == 1:
                cls_name, cfg = next(iter(node.items()))
                cls = subclasses.get(cls_name, base_cls)
                # Determine unique child name
                count = name_counts.get(cls_name, 0) + 1
                name_counts[cls_name] = count
                child_name = cls_name if count == 1 else f"{cls_name}_{count - 1}"
                cls(name=child_name, parent=self, definition=cfg)
            elif isinstance(node, dict) and len(node) > 1:
                it = iter(node.items())
                cls_name, cfg = next(it)
                cls = subclasses.get(cls_name, base_cls)
                # Determine unique root name
                count = name_counts.get(cls_name, 0) + 1
                name_counts[cls_name] = count
                root_name = cls_name if count == 1 else f"{cls_name}_{count - 1}"
                root = cls(name=root_name, parent=self, definition=cfg)
                for sub_name, sub_cfg in it:
                    sub_cls = subclasses.get(sub_name, base_cls)
                    # For sub-children, no need to index as keys are unique under this root
                    sub_cls(name=sub_name, parent=root, definition=sub_cfg)
            else:
                # Scalar or unsupported dict form
                name = base_cls.__name__ if not isinstance(node, str) else node
                cfg = None if isinstance(node, str) else node
                # Determine unique child name
                count = name_counts.get(name, 0) + 1
                name_counts[name] = count
                child_name = name if count == 1 else f"{name}_{count - 1}"
                base_cls(name=child_name, parent=self, definition=cfg)

    def _handle_scalar_filtered(self, definition: Any, base_name: str, base_cls):
        """Instantiate a single child for scalar filtered definitions."""
        child = base_cls(name=base_name, parent=self, definition=definition)
        self.add_child(child)

    # Filtered mode: instantiate children using the specified type filter

    def _load_filtered(self, definition: Any) -> None:
        base_name = self._type.__name__
        subclasses = get_classes_by_base(base_name)
        base_entry = class_registry.get(base_name)
        if not base_entry:
            raise ValueError(f"Base class {base_name} not found in registry")
        base_cls = base_entry["class"]

        if isinstance(definition, dict):
            self._handle_dict_filtered(definition, base_cls, subclasses)
        elif isinstance(definition, list):
            self._handle_list_filtered(definition, base_cls, subclasses)
        else:
            self._handle_scalar_filtered(definition, base_name, base_cls)

    # Generic mode: instantiate children by registry lookup without type filter

    def _handle_dict_generic(self, definition: dict) -> None:
        """Instantiate children for generic dict definitions."""
        # Iterate over a static list of keys to avoid modifying while iterating
        for cls_name in list(definition.keys()):
            if cls_name not in class_registry:
                continue
            cfg = definition.pop(cls_name)
            create_instance(
                cls_name,
                name=cls_name,
                parent=self,
                definition=cfg
            )

    def _handle_list_generic(self, definition: list) -> None:
        """Instantiate children for generic list definitions."""
        # Process each element by popping from the front until empty,
        # ensure unique child names for duplicate class entries
        name_counts: dict[str, int] = {}
        original_len = len(definition)
        for _ in range(original_len):
            node = definition.pop(0)
            if isinstance(node, dict) and len(node) == 1:
                cls_name, cfg = next(iter(node.items()))
                if cls_name not in class_registry:
                    raise ValueError(f"Class '{cls_name}' not found in registry")
                # determine unique child name
                count = name_counts.get(cls_name, 0) + 1
                name_counts[cls_name] = count
                child_name = cls_name if count == 1 else f"{cls_name}_{count - 1}"
                create_instance(
                    cls_name,
                    name=child_name,
                    parent=self,
                    definition=cfg
                )
            elif isinstance(node, dict) and len(node) > 1:
                it = iter(node.items())
                cls_name, cfg = next(it)
                if cls_name not in class_registry:
                    raise ValueError(f"Class '{cls_name}' not found in registry")
                # determine unique root name
                count = name_counts.get(cls_name, 0) + 1
                name_counts[cls_name] = count
                root_name = cls_name if count == 1 else f"{cls_name}_{count - 1}"
                root = create_instance(
                    cls_name,
                    name=root_name,
                    parent=self,
                    definition=cfg
                )
                for sub_name, sub_cfg in it:
                    if sub_name not in class_registry:
                        raise ValueError(f"Class '{sub_name}' not found in registry")
                    # direct child under root; use sub_name (assuming no duplicates at this level)
                    create_instance(
                        sub_name,
                        name=sub_name,
                        parent=root,
                        definition=sub_cfg
                    )
            # ignore non-dict nodes

    def _load_generic(self, definition: Any) -> None:
        """
        Generic loading mode: treat each entry in a dict definition as
        class_name → configuration, or each node in a list as a single-key dict.
        Raises ValueError if a referenced class is not registered.
        """
        if isinstance(definition, dict):
            self._handle_dict_generic(definition)
        elif isinstance(definition, list):
            self._handle_list_generic(definition)
        else:
            # Scalar definitions are not loaded in generic mode
            return



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from .attribute import type_mapping, SpxAttribute, InternalAttributeWrapper, ExternalAttributeWrapper, StaticAttributeWrapper
from .attributes import SpxAttributes
from spx_sdk.attributes.resolve_attribute import (
    is_attribute_reference,
    find_attribute,
    resolve_attribute_reference,
    get_attribute_value,
    set_attribute_value,
    resolve_reference,
    resolve_attribute_reference_hierarchical,
    substitute_attribute_references_hierarchical,
    extract_attribute_wrappers_hierarchical,
    substitute_with_wrappers,
    resolve_nested_chain_reference,
)

__all__ = ["SpxAttribute",
           "InternalAttributeWrapper",
           "ExternalAttributeWrapper",
           "StaticAttributeWrapper",
           "SpxAttributes",
           "type_mapping",
           "is_attribute_reference",
           "find_attribute",
           "resolve_attribute_reference",
           "get_attribute_value",
           "set_attribute_value",
           "resolve_reference",
           "resolve_attribute_reference_hierarchical",
           "substitute_attribute_references_hierarchical",
           "extract_attribute_wrappers_hierarchical",
           "substitute_with_wrappers",
           "resolve_nested_chain_reference",
           ]



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

import time
from typing import Dict, Any

from spx_sdk.registry import register_class
from spx_sdk.components import SpxComponent


@register_class(name="simple_timer")
class SimpleTimer(SpxComponent):
    """SimpleTimer component that measures elapsed time with configurable resolution."""

    def _populate(self, definition: Dict[str, Any]) -> None:
        self.resolution = definition.get("resolution", None)
        self.reset()

    def reset(self) -> bool:
        self.running = False
        self.start_time = 0.0
        self.timer = 0.0
        return True

    def start(self) -> bool:
        if not self.running:
            self.start_time = time.time()
            self.running = True
            return True
        return False

    def stop(self) -> bool:
        if self.running:
            self.timer = time.time() - self.start_time
            self.running = False
            return True
        return False

    def elapsed(self) -> float:
        if self.running:
            return time.time() - self.start_time
        return self.timer

    def is_running(self) -> bool:
        return self.running

    @property
    def time(self) -> float:
        elapsed = self.elapsed()
        if self.resolution:
            elapsed = round(elapsed / self.resolution) * self.resolution
        return elapsed

    @time.setter
    def time(self, value: float) -> None:
        if not isinstance(value, (int, float)):
            raise ValueError("time must be a numeric value")
        if self.running:
            self.stop()
        self.timer = float(value)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from spx_sdk.hooks.hooks import Hooks
from spx_sdk.hooks.refresh import RefreshHook

__all__ = [
    "Hooks",
    "RefreshHook"
]



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from spx_sdk.components import SpxComponent


class Protocol(SpxComponent):
    """
    Protocol class for managing protocol components.
    Inherits from SpxContainer to manage a collection of components.
    """
    pass



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from spx_sdk.registry import register_class, load_module_from_path
from spx_sdk.components import SpxComponent
from spx_sdk.attributes import SpxAttribute


@register_class(name="python_file")
@register_class(name="import")
class PythonFile(SpxComponent):

    def _populate(self, definition: dict) -> None:
        self.class_instances = {}
        for module_path, params in definition.items():
            params["path"] = module_path
            class_instance = self.create_instance_from_module(params)
            class_name = params["class"]
            self.class_instances[class_name] = class_instance

    def create_instance_from_module(self, module_info: dict):
        file_path = module_info["path"]
        class_name = module_info["class"]
        module = load_module_from_path(file_path)
        cls = getattr(module, class_name)

        # Extract custom init parameters if provided
        init_info = module_info.get("init", {})
        init_args = init_info.get("args", [])
        init_kwargs = init_info.get("kwargs", {})

        if SpxComponent in cls.__bases__:
            # Prepend root and definition for Item subclasses
            args = [self.get_root(), self.definition] + init_args
            class_instance = cls(*args, **init_kwargs)
        else:
            # Instantiate plain classes with provided args/kwargs
            class_instance = cls(*init_args, **init_kwargs)
        return class_instance

    def prepare(self):
        if isinstance(self.definition, dict):
            for module_path, params in self.definition.items():
                for attr, methods in params["attributes"].items():
                    class_name = params["class"]
                    attribute: SpxAttribute = self.get_root().get("attributes").get(attr)
                    if "property" in methods:
                        attribute.link_to_internal_property(self.class_instances[class_name], methods["property"])
                    elif "getter" in methods:
                        getter_name = methods["getter"]
                        setter_name = methods.get("setter", None)
                        attribute.link_to_internal_func(self.class_instances[class_name], getter_name, setter_name)

    def reset(self):
        for module_path, params in self.definition.items():
            for attr, methods in params["attributes"].items():
                attribute: SpxAttribute = self.get_root().get("attributes").get(attr)
                attribute.unlink_internal_property()



from typing import Any, Dict
from spx_sdk.registry import register_class
from spx_sdk.actions.action import Action


@register_class(name="set")
@register_class(name="Set")
class SetAction(Action):
    """
    SetAction class for assigning a literal value to one or more attributes.
    Inherits from Action to manage outputs and inputs.
    """

    def _populate(self, definition: Dict[str, Any]) -> None:
        """
        Populate the SetAction with the given definition.
        Extracts 'value' from definition.params and
        then delegates to base class.
        """
        # Extract literal value: can be top‑level or inside "params"
        if "value" in definition:
            self.value = definition.pop("value")
            params = definition.get("params", {})
        else:
            params = definition.get("params", {})
            self.value = params.pop("value", None)
        # Now let base Action handle function, outputs, and remaining params
        super()._populate(definition)

    def run(self, *args, **kwargs) -> Any:
        """
        Execute the set operation: assign the literal value to all outputs.
        Returns the value that was set.
        """
        # Assign the stored literal value to each output wrapper
        for wrapper in self.outputs.values():
            wrapper.set(self.value)
        return True



from typing import Any
from spx_sdk.registry import register_class
from spx_sdk.actions.action import Action


@register_class(name="function")
class FunctionAction(Action):
    """
    FunctionAction class for executing a function.
    Inherits from Action to manage action components.
    """

    def _populate(self, definition: dict) -> None:
        self.call = None
        super()._populate(definition)

    def run(self, *args, **kwargs) -> Any:
        """
        Evaluate the call expression, resolving attribute references,
        and write the result to all output attributes.
        """
        super().run()
        if self.call is None:
            return False  # No call defined, nothing to run
        return self.write_outputs(self.call)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from typing import Any

from spx_sdk.registry import register_class, create_instance
from spx_sdk.components import SpxContainer


@register_class(name="hooks")
@register_class(name="Hooks")
class Hooks(SpxContainer):
    """
    SPX container for hook definitions. Each hook entry is turned into a Hook child.

    Definition must be a dict mapping hook names to:
      - a single attribute reference string, or
      - a list of attribute reference strings.
    """
    def _populate(self, definition: Any) -> None:
        if not isinstance(definition, dict):
            raise ValueError("Hooks definition must be a dict mapping hook names to references")
        for event_name, hook_entries in definition.items():
            # Normalize entries to a list
            if isinstance(hook_entries, (str, dict)):
                entries = [hook_entries]
            elif isinstance(hook_entries, list):
                entries = hook_entries
            else:
                raise ValueError(
                    f"Invalid hook definition for '{event_name}': expected string, dict, or list"
                )

            for idx, entry in enumerate(entries):
                # Determine class name and its specific definition
                if isinstance(entry, str):
                    cls_name = entry
                    entry_def = {}
                elif isinstance(entry, dict) and len(entry) == 1:
                    cls_name, entry_def = next(iter(entry.items()))
                else:
                    raise ValueError(
                        f"Invalid hook entry for '{event_name}': {entry!r}"
                    )

                # Unique instance name: event + class + optional index
                inst_name = cls_name if idx == 0 else f"{cls_name}_{idx}"
                # Instantiate the registered hook class under this container
                hook_inst = create_instance(
                    cls_name,
                    name=inst_name,
                    parent=self,
                    definition=entry_def
                )
                # Register this hook component on the parent under its event name
                if hasattr(self.parent, "register_hook"):
                    self.parent.register_hook(event_name, hook_inst)

    def run(self, event_name: str, *args, **kwargs) -> None:
        return True



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from typing import Any, Dict, Optional
from collections.abc import MutableMapping
from spx_sdk.components import SpxContainer, SpxComponent
from spx_sdk.attributes import SpxAttribute
from spx_sdk.registry import register_class


class InternalValue(MutableMapping):
    """
    Mutable mapping view over attribute values.

    Keys are attribute names, values are the respective internal/external values.
    """
    def __init__(self, attributes: Dict[str, SpxAttribute]) -> None:
        self._attributes = attributes

    def __setitem__(self, name: str, value: Any) -> None:
        # If the attribute exists, update its internal_value
        if name in self._attributes:
            self._attributes[name].internal_value = value
        else:
            raise KeyError(f"Attribute {name} did not exists.")

    def __getitem__(self, name: str):
        # Retrieve the internal_value for the named attribute
        if name in self._attributes:
            return self._attributes[name].internal_value
        else:
            raise KeyError(f"Attribute {name} did not exists.")

    def __delitem__(self, key: str) -> None:
        """Deletion is not supported."""
        raise NotImplementedError("Cannot delete items from InternalValue.")

    def __iter__(self):
        """Iterate over attribute names."""
        return iter(self._attributes)

    def __len__(self) -> int:
        """Return number of attributes."""
        return len(self._attributes)


class ExternalValue(MutableMapping):
    """
    Mutable mapping view over attribute values.

    Keys are attribute names, values are the respective internal/external values.
    """
    def __init__(self, attributes: Dict[str, SpxAttribute]) -> None:
        self._attributes = attributes

    def __setitem__(self, name: str, value: Any) -> None:
        # If the attribute exists, update its external_value
        if name in self._attributes:
            self._attributes[name].external_value = value
        else:
            raise KeyError(f"Attribute {name} did not exists.")

    def __getitem__(self, name: str):
        # Retrieve the external_value for the named attribute
        if name in self._attributes:
            return self._attributes[name].external_value
        else:
            raise KeyError(f"Attribute {name} did not exists.")

    def __delitem__(self, key: str) -> None:
        """Deletion is not supported."""
        raise NotImplementedError("Cannot delete items from ExternalValue.")

    def __iter__(self):
        """Iterate over attribute names."""
        return iter(self._attributes)

    def __len__(self) -> int:
        """Return number of attributes."""
        return len(self._attributes)


@register_class(name="Attributes")
@register_class(name="attributes")
class SpxAttributes(SpxContainer):
    """
    Container for SpxAttribute instances, providing direct value access.

    Attributes:
        children (Dict[str, SpxAttribute]): Mapping of attribute names to SpxAttribute instances.
        internal (_ValueView): View for getting/setting internal values.
        external (_ValueView): View for getting/setting external values.
    """

    def __init__(
        self,
        name: str,
        definition: Any,
        *,
        parent: Optional[SpxComponent] = None,
    ):
        # Tell SpxContainer: "load these entries, filtered to type=Attribute"
        super().__init__(name=name, definition=definition, parent=parent, type=SpxAttribute)

    def _populate(self, definition):
        super()._populate(definition)
        # Now SpxContainer has created one child per dict‐entry, each an Attribute.
        # And wire up our two “views”:
        self.internal = InternalValue(self.children)
        self.external = ExternalValue(self.children)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

import uuid
from spx_sdk.registry import register_class
from spx_sdk.components import SpxContainer
from typing import Any, Optional
from spx_sdk.components import SpxComponent
from spx_sdk.model.timer import SimpleTimer
from spx_sdk.model.polling import SimplePolling


@register_class(name="base_model")
class BaseModel(SpxContainer):
    """
    Base SPX model container. Each instance has a unique `uid`.
    """
    def __init__(
        self,
        name: Optional[str] = None,
        definition: Any = None,
        parent: Optional[SpxComponent] = None
    ):
        # Assign a unique identifier for this model instance
        self.uid = str(uuid.uuid4())

        # Determine model definition dict or use empty
        if isinstance(definition, dict):
            model_def = definition
        else:
            model_def = {}

        # Extract (and remove) any nested 'timer' or 'polling' entries
        timer_def = model_def.pop("timer", {}) or {}
        polling_def = model_def.pop("polling", {}) or {}

        # Initialize base container with the remaining definition
        super().__init__(name=name, definition=model_def, parent=parent, type=SpxContainer)

        # Instantiate Timer and Polling children using their extracted definitions
        SimpleTimer(name="timer", parent=self, definition=timer_def)
        SimplePolling(name="polling", parent=self, definition=polling_def)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from spx_sdk.components import SpxContainer
from spx_sdk.registry import register_class

"""
type_mapping: dict
    A mapping from data type names (string) to a dict specifying the Python
    'type' and a 'default' value for that type. Supported keys are:
    "float", "bool", "int", and "str".
"""
type_mapping = {
    "float": {"type": float, "default": 0.0},
    "bool": {"type": bool, "default": False},
    "int": {"type": int, "default": 0},
    "str": {"type": str, "default": ""}
}


class LinkedProperty:
    """
    A helper that binds a named getter and optional setter on a target instance
    so that getting or setting through this object delegates to the underlying
    attribute or method of that instance.
    """
    def __init__(self, instance, getter_name, setter_name=None):
        self._instance = instance
        self._getter_name = getter_name
        if not setter_name:
            setter_name = getter_name
        self._setter_name = setter_name

    def get_value(self):
        if self._instance is None:
            raise AttributeError("Linked property is not set.")
        getter = getattr(self._instance, self._getter_name)

        if callable(getter):
            return getter()
        else:
            return getter

    def set_value(self, value):
        if self._instance is None:
            raise AttributeError("Linked property is not set.")
        setter = getattr(type(self._instance), self._setter_name)
        if callable(setter):
            setter(self._instance, value)
        elif isinstance(setter, property):
            if setter.fset:
                setattr(self._instance, self._setter_name, value)
        else:
            raise AttributeError(f"{self._setter_name} is neither a function nor a property")


@register_class(name="Attribute")
@register_class(name="attribute")
@register_class()
class SpxAttribute(SpxContainer):
    """
    Represents a typed model attribute with separate internal and external values,
    trigger callbacks on value changes, and the ability to link its values to
    properties or methods of its owning component.
    """

    def _populate(self, definition):
        """
        Populate the attribute from a definition dict or a single value.
        If a dict is provided, it can contain:
        - 'type': the data type of the attribute (default is 'float')
        - 'default': the initial value for the attribute (optional)
        If a single value is provided, it is treated as the initial value.
        """

        if not isinstance(definition, dict):
            self.type = type(definition).__name__
            if self.type not in type_mapping:
                raise ValueError(f"Unsupported type: {self.type}. Supported types are: {', '.join(type_mapping.keys())}")
            self.initial_value = definition
        else:
            # Infer type: explicit 'type' key or derive from default value
            if "type" in definition:
                t = definition["type"]
            elif "default" in definition:
                default_val = definition["default"]
                t = type(default_val).__name__
            else:
                t = "float"

            # Validate against supported types
            if t not in type_mapping:
                raise ValueError(
                    f"Unsupported type: {t}. Supported types are: {', '.join(type_mapping.keys())}"
                )
            self.type = t
            # Set initial value to provided default or mapping default
            self.initial_value = definition.get(
                "default", type_mapping[self.type]["default"]
            )
            definition.pop("type", None)  # Remove type from definition to avoid conflicts
            definition.pop("default", None)  # Remove default from definition to avoid conflicts

        self._external_value = None
        self._internal_value = self.initial_value
        self._linked_internal_property = None
        self._linked_external_property = None
        self.unit = None

        super()._populate(definition)

    @property
    def external_value(self):
        if self._external_value is not None:
            return self._external_value
        return self._internal_value

    @external_value.setter
    def external_value(self, value):
        expected_type = type_mapping.get(self.type, type_mapping["str"])["type"]
        try:
            if self._linked_external_property:
                self._linked_external_property.set_value(expected_type(value))
            self._external_value = expected_type(value)  # Change of external value will set external value
            # Trigger hooks for value change
            self.trigger_hooks("on_set")
            self.trigger_hooks("on_external_set")
        except (ValueError, TypeError) as e:
            raise TypeError(f"Expected type {self.type}, but given {type(value).__name__}") from e

    @property
    def internal_value(self):
        if self._linked_internal_property:
            self._internal_value = self._linked_internal_property.get_value()
        return self._internal_value

    @internal_value.setter
    def internal_value(self, value):
        expected_type = type_mapping.get(self.type, type_mapping["str"])["type"]
        try:
            if self._linked_internal_property:
                self._linked_internal_property.set_value(expected_type(value))
            self._internal_value = expected_type(value)
            self._external_value = None  # Change of internal value will reset external value
            # Trigger hooks for value change
            self.trigger_hooks("on_set")
            self.trigger_hooks("on_internal_set")
        except (ValueError, TypeError) as e:
            raise TypeError(f"Expected type {self.type}, but given {type(value).__name__}") from e

    def link_to_internal_property(self, instance, property_name):
        internal_value = self.internal_value
        self._linked_internal_property = LinkedProperty(instance, property_name)
        self._linked_internal_property.set_value(internal_value)

    def link_to_external_property(self, instance, property_name):
        external_value = self.external_value
        self._linked_external_property = LinkedProperty(instance, property_name)
        self._linked_external_property.set_value(external_value)

    def link_to_internal_func(self, instance, getter, setter=None):
        internal_value = self.internal_value
        self._linked_internal_property = LinkedProperty(instance, getter, setter)
        if setter:
            self._linked_internal_property.set_value(internal_value)

    def link_to_external_func(self, instance, getter, setter=None):
        external_value = self.external_value
        self._linked_external_property = LinkedProperty(instance, getter, setter)
        if setter:
            self._linked_external_property.set_value(external_value)

    def unlink_internal_property(self):
        self._linked_internal_property = None

    def unlink_external_property(self):
        self._linked_external_property = None

    def prepare(self):
        self._external_value = None

    def run(self):
        self._external_value = None

    @property
    def internal(self):
        """
        Returns a wrapper to get/set this attribute's internal value.
        """
        return InternalAttributeWrapper(self)

    @property
    def external(self):
        """
        Returns a wrapper to get/set this attribute's external value.
        """
        return ExternalAttributeWrapper(self)

    def get_wrapper(self, wrapper_type: str) -> 'InternalAttributeWrapper | ExternalAttributeWrapper':
        """
        Return the corresponding attribute wrapper by name.

        Args:
            wrapper_type (str): 'internal' or 'external'.

        Returns:
            InternalAttribute or ExternalAttribute wrapper instance.
        """
        t = wrapper_type.lower()
        if t == "internal":
            return InternalAttributeWrapper(self)
        if t == "external":
            return ExternalAttributeWrapper(self)
        raise ValueError(f"Unknown wrapper type '{wrapper_type}'")

    def get(self):
        """Return the internal value."""
        return self.internal_value

    def set(self, value):
        """Set the internal value."""
        self.internal_value = value

    def __getitem__(self, key: str):
        # Leaf: try to return attribute
        if hasattr(self, key):
            return getattr(self, key)
        raise KeyError(f"Attribute '{self.name}' has no child or attribute named '{key}'.")

    def __setitem__(self, key: str, value) -> None:
        # Leaf: attempt to set property or attribute
        cls_attr = getattr(self.__class__, key, None)
        # If it's a property on the class
        if isinstance(cls_attr, property):
            if cls_attr.fset is None:
                raise AttributeError(f"Property '{key}' on attribute '{self.name}' is read-only")
            # Use setattr to invoke the property's setter
            setattr(self, key, value)
        else:
            # Not a property: set or create a normal attribute
            setattr(self, key, value)


class InternalAttributeWrapper:
    """
    Wrapper for accessing and setting the internal_value of an SpxAttribute.
    """
    def __init__(self, attribute: SpxAttribute):
        self._attribute = attribute
        self.name = attribute.name
        self.type = attribute.type

    def get(self):
        """Return the internal value."""
        return self._attribute.internal_value

    def set(self, value):
        """Set the internal value."""
        self._attribute.internal_value = value

    def __repr__(self):
        return f"<Internal {self._attribute.name}={self.get()!r}>"


class ExternalAttributeWrapper:
    """
    Wrapper for accessing and setting the external_value of an SpxAttribute.
    """
    def __init__(self, attribute: SpxAttribute):
        self._attribute = attribute
        self.name = attribute.name
        self.type = attribute.type

    def get(self):
        """Return the external value."""
        return self._attribute.external_value

    def set(self, value):
        """Set the external value."""
        self._attribute.external_value = value

    def __repr__(self):
        return f"<External {self._attribute.name}={self.get()!r}>"


# New wrapper for plain Python attributes
class StaticAttributeWrapper:
    """
    Wrapper for accessing and setting a plain Python attribute on any object.
    """
    def __init__(self, instance, attr_name):
        self._instance = instance
        self.name = attr_name

    def get(self):
        """Return the attribute's current value."""
        return getattr(self._instance, self.name)

    def set(self, value):
        """Set the attribute's value."""
        setattr(self._instance, self.name, value)

    def __repr__(self):
        val = self.get()
        return f"<Static {self.name}={val!r}>"



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from .component import SpxComponent, SpxComponentState
from .containter import SpxContainer

__all__ = ["SpxComponent",
           "SpxComponentState",
           "SpxContainer"
           ]



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

__version__ = "0.1.0"

import pkgutil
import importlib

# Dynamically import all submodules so that @register_class decorators run
for _, mod_name, _ in pkgutil.iter_modules(__path__):
    importlib.import_module(f"{__name__}.{mod_name}")

__all__ = [
    "__version__",
]



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

import importlib
import os
import sys
import subprocess
import logging
import fnmatch
import yaml
from copy import deepcopy

class_registry = {}
instance_registry = {}


def register_class(cls=None, *, name=None):
    """
    Register a class in the registry.
    If a name is provided, it will be used as the key in the registry.
    If no name is provided, the class's __name__ will be used.
    If cls is None, return a decorator that registers the class when called.
    :param cls: The class to register.
    :param name: Optional name for the class in the registry.
                 If not provided, the class's __name__ will be used.
    :return: The class itself.
    """
    if cls is None:
        return lambda cls: register_class(cls, name=name)

    class_name = name if name else cls.__name__
    base_class = cls.__bases__[0].__name__
    class_registry[class_name] = {"class": cls, "base_class": base_class}
    return cls


def create_instance(class_name, *args, **kwargs):
    """
    Create an instance of a registered class or YAML template by its name.

    If the registry entry includes a 'template' dict, merges that template
    with any provided 'definition' in kwargs (caller’s values override
    the template). Then calls the constructor of the registered class
    (often SpxContainer) with the merged definition.

    If the class_name is not in the registry, attempts a dynamic import of
    a Python class path. Raises ValueError if no class can be found.

    Args:
        class_name (str): Name of the class or template to instantiate.
        *args: Positional arguments forwarded to the class constructor.
        **kwargs: Keyword arguments forwarded to the class constructor,
                  including an optional 'definition' dict.

    Returns:
        object: New instance of the target class.

    Raises:
        ValueError: If the class_name is not registered and cannot be imported.
    """
    entry = class_registry.get(class_name)
    if entry:
        cls = entry["class"]
        # If this entry carries a YAML‐template, merge it with the caller’s definition
        if "template" in entry:
            # grab the template dict
            tmpl = entry["template"] or {}
            # grab the caller’s definition (e.g. parameters) or empty dict
            own_def = kwargs.get("definition", {}) or {}
            # shallow‐merge (caller’s keys override template’s)
            merged = {**tmpl, **own_def}
            kwargs["definition"] = deepcopy(merged)
    else:
        # fallback: try to import a real Python class
        cls = dynamic_import(class_name)
        if cls is None:
            raise ValueError(f"Class {class_name} not found and could not be imported")

    # Finally call the initializer
    return cls(*args, **kwargs)


def dynamic_import(class_path):
    """
    Dynamically import a class from a module.
    The class_path should be in the format 'module.submodule.ClassName'.
    This function will attempt to import the module and then get the class from it.
    If the class cannot be found, it will return None.
    :param class_path: The full path to the class (e.g., 'module.submodule.ClassName').
    :return: The class if found, None otherwise.
    """
    try:
        module_path, class_name = class_path.rsplit(".", 1)
        module = importlib.import_module(module_path)
        cls = getattr(module, class_name)
        return cls
    except (ImportError, AttributeError, ValueError) as e:
        print(f"Error importing {class_path}: {e}")
        return None


def load_module_from_path(file_path, module_name=None):
    """
    Load a module from a given file path.
    If module_name is provided, it will be used as the module name.
    If not, the module name will be derived from the file name.
    This function will remove the module from sys.modules if it already exists,
    allowing for reloading the module.
    :param file_path: Path to the .py file.
    :param module_name: Optional name for the module.
                        If not provided, the file name (without extension) will be used.
    :return: The loaded module.
    """
    if not module_name:
        module_name = os.path.splitext(os.path.basename(file_path))[0]
    # If the module exists in sys.modules, remove it
    if module_name in sys.modules:
        del sys.modules[module_name]
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def load_modules_from_directory(directory, skip_pattern="*test*"):
    """
    Recursively traverse the given directory and load all .py modules (except __init__.py),
    skipping files whose names match the specified skip_pattern.
    :param directory: Base directory to start loading modules from.
    :param skip_pattern: Pattern used to filter out files.
                         Default is "*test*", which will skip any file whose name contains "test" (case-insensitive).
    """
    if directory is None:
        return
    directory_path = os.path.abspath(directory)
    if directory_path not in sys.path:
        sys.path.append(directory_path)

    for filename in os.listdir(directory_path):
        # If the file's name matches skip_pattern, skip it.
        if fnmatch.fnmatch(filename.lower(), skip_pattern.lower()):
            continue
        if filename.endswith(".py") and filename != "__init__.py":
            logging.debug("Loading module: %s", filename)
            module_name = os.path.splitext(filename)[0]
            load_module_from_path(os.path.join(directory_path, filename), module_name)


def load_modules_recursively(directory, skip_pattern="*test*"):
    """
    Recursively traverse the given directory and load all .py modules (except __init__.py),
    while skipping any directories or files whose name matches the skip_pattern
    (case-insensitive).
    :param directory: The base directory to traverse.
    :param skip_pattern: Pattern to use for skipping directories or files.
                         Default is "*test*", so any folder or file containing "test" (any case) is skipped.
    """
    if directory is None:
        return

    directory_path = os.path.abspath(directory)
    if directory_path not in sys.path:
        sys.path.append(directory_path)

    for root, dirs, files in os.walk(directory_path):
        # Exclude subdirectories whose name matches the skip pattern.
        dirs[:] = [d for d in dirs if not fnmatch.fnmatch(d.lower(), skip_pattern.lower())]
        if root not in sys.path:
            sys.path.append(root)
        for filename in files:
            # Skip files matching the skip pattern.
            if fnmatch.fnmatch(filename.lower(), skip_pattern.lower()):
                continue
            if filename.endswith(".py") and filename != "__init__.py":
                filepath = os.path.join(root, filename)
                # Create a module name based on the file's relative path from the base directory.
                rel_path = os.path.relpath(filepath, directory_path)
                module_name = os.path.splitext(rel_path)[0].replace(os.sep, ".")
                logging.debug("Loading module %s from path %s", module_name, filepath)
                load_module_from_path(filepath, module_name)


def install_requirements_from_dir(directory, requirements_pattern="requirements*.txt"):
    """
    Recursively traverse the given directory. For each file that matches the given
    requirements_pattern (default: "requirements*.txt"), install its packages using pip.
    Uses the PIP_INDEX_URL environment variable if set (pointing to a local mirror).
    :param directory: The base directory to traverse.
    :param requirements_pattern: Pattern to match requirement files.
                                 Default is "requirements*.txt".
    :return: None
    """
    for root, dirs, files in os.walk(directory):
        for file in files:
            # Sprawdzamy, czy nazwa pliku pasuje do wzorca, zignorowaliśmy wielkość liter
            if fnmatch.fnmatch(file.lower(), requirements_pattern.lower()):
                req_file = os.path.join(root, file)
                print(f"Installing requirements from {req_file}...")
                cmd = [sys.executable, "-m", "pip", "install", "--no-cache-dir", "-r", req_file]
                local_index = os.environ.get("PIP_INDEX_URL")
                if local_index:
                    cmd.extend(["--index-url", local_index])
                subprocess.check_call(cmd)


def get_classes_by_base(base_class_name: str):
    """
    Return all registered classes that inherit (directly or indirectly) from the given base class name.
    :param base_class_name: The name of the base class to filter by.
    :return: A dict mapping class names to class objects for all subclasses of base_class_name.
    """
    result = {}
    for name, info in class_registry.items():
        cls = info["class"]
        # Check all ancestor classes in MRO (excluding the class itself)
        ancestor_names = [c.__name__ for c in cls.__mro__[1:]]
        if base_class_name in ancestor_names:
            result[name] = cls
    return result


def get_class_names_by_base(base_class_name: str):
    """
    Return all registered class names whose recorded base_class matches base_class_name.
    :param base_class_name: The name of the base class to filter by.
    :return: A list of class names that inherit from base_class_name.
    """
    return [
        name
        for name, info in class_registry.items()
        if info["base_class"] == base_class_name
    ]


def filter_instances_by_base_class(base_class):
    """
    Filter instances by base class.
    This function returns a dictionary of instance names and their corresponding instances
    that are instances of the specified base class.
    The base class must be registered in the class_registry.
    :param base_class: The base class to filter by.
    :return: A dictionary of instance names and their corresponding instances.
    """
    filtered_instances = {}
    for instance_name, instance in instance_registry.items():
        if isinstance(instance, base_class):
            filtered_instances[instance_name] = instance
    return filtered_instances


def filter_instances_by_base_class_name(base_class_name):
    """
    Filter instances by base class name.
    This function returns a dictionary of instance names and their corresponding instances
    that are instances of the specified base class name.
    The base class name must be registered in the class_registry.
    :param base_class_name: The base class name to filter by.
    :return: A dictionary of instance names and their corresponding instances.
    """
    filtered_instances = {}
    for instance_name, instance in instance_registry.items():
        for instance_base_class in type(instance).__bases__:
            if instance_base_class.__name__ == base_class_name:
                filtered_instances[instance_name] = instance
    return filtered_instances


def get_instance(instance_name):
    """
    Get an instance by name.
    The instance must be registered in the instance_registry.
    :param instance_name: The name of the instance to retrieve.
    :return: The instance if found, None otherwise.
    """
    if instance_name in instance_registry:
        return instance_registry[instance_name]
    else:
        raise ValueError(f"Instance {instance_name} not found in registry.")


def get_all_instances():
    """
    Get all instances in the instance registry.
    :return: A dictionary of instance names and their corresponding instances.
    """
    if instance_registry is None:
        raise ValueError("Instance registry is empty.")
    return instance_registry


def get_all_classes():
    """
    Get all classes in the class registry.
    :return: A dictionary of class names and their corresponding classes.
    """
    if class_registry is None:
        raise ValueError("Class registry is empty.")
    return class_registry


def clear_registry():
    """
    Clear the class and instance registries.
    This function resets the class_registry and instance_registry to empty dictionaries.
    :return: None
    """
    class_registry.clear()
    instance_registry.clear()
    logging.debug("Cleared class and instance registries.")


def get_class(class_name: str):
    """
    Get class by name.
    The class must be registered in the class_registry.
    :param class_name: The name of the class to retrieve.
    :return: The class if found, None otherwise.
    """
    if class_name not in class_registry:
        raise ValueError(f"Class {class_name} not found in registry.")
    return class_registry[class_name]["class"]


def get_class_base(class_name: str):
    """
    Get base class name by class name.
    The class must be registered in the class_registry.
    :param class_name: The name of the class to retrieve.
    :return: The base class name if found, None otherwise.
    """
    if class_name not in class_registry:
        raise ValueError(f"Class {class_name} not found in registry.")
    return class_registry[class_name]["base_class"]


def load_instances_from_yaml(filename: str):
    """
    Load instances from a YAML file.
    The YAML file should contain a mapping of instance names to their class names and parameters.
    The class must be registered in the class_registry.
    :param filename: Path to the YAML file.
    """
    if not os.path.exists(filename):
        raise FileNotFoundError(f"File {filename} not found.")
    with open(filename, "r") as file:
        load_instances_from_yaml_data(file)


def load_instances_from_yaml_data(yaml_data: str):
    """
    Load instances from a YAML string.
    The YAML string should contain a mapping of instance names to their class names and parameters.
    The class must be registered in the class_registry.
    :param yaml_data: YAML string containing instance definitions.
    """
    data = yaml.safe_load(yaml_data)

    for instance_name, instance_info in data.items():
        class_name = instance_info["class"]
        parameters = instance_info.get("parameters", {})
        instance = create_instance(class_name, **parameters)
        instance_registry[instance_name] = instance
        logging.debug(f"Loaded instance {instance_name} of class {class_name} with parameters {parameters}")



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from .model_base import BaseModel
Model = BaseModel
from .polling import SimplePolling
from .timer import SimpleTimer


__all__ = ["BaseModel",
           "Model",
           "SimplePolling",
           "SimpleTimer",
           ]



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

import logging
import dataclasses
from spx_sdk.registry import register_class
from typing import Optional, Any, List, Dict
from enum import Enum, auto


class SpxComponentState(Enum):
    INITIALIZED = auto()
    PREPARING = auto()
    PREPARED = auto()
    STARTING = auto()
    STARTED = auto()
    RUNNING = auto()
    PAUSING = auto()
    PAUSED = auto()
    STOPPING = auto()
    STOPPED = auto()
    DESTROYING = auto()
    DESTROYED = auto()
    RESETTING = auto()
    RESET = auto()
    FAULT = auto()
    UNKNOWN = auto()

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


@register_class()
class SpxComponent:
    """
    SpxComponent represents a node in a hierarchical component tree.

    Attributes:
        name (str): Unique name of this component instance.
        parent (Optional[SpxComponent]): Parent component in the hierarchy.
        children (Dict[str, SpxComponent]): Mapping of child component names to instances.
        definition (Any): Configuration data used to populate this component.
        state (SpxComponentState): Current lifecycle state of the component.

    This class provides methods to manage child components, control the component’s lifecycle
    (prepare, run, start, pause, stop, reset, destroy), and navigate the component tree.
    It supports dict-like access (`component['child']`), containment checks
    (`'child' in component`), and length queries (`len(component)`).
    """
    def __init__(
            self,
            name: Optional[str] = None,
            parent: Optional["SpxComponent"] = None,
            definition: Optional[Any] = None
            ):
        self.name = name if name is not None else self.__class__.__name__
        self.parent = parent
        self.definition: Any = definition
        self.children: Dict[str, SpxComponent] = {}
        self.logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}.{self.name}"
        )
        if parent is not None:
            parent.add_child(self)

        self.state = SpxComponentState.INITIALIZED
        # Component enabled flag: when False, lifecycle methods do nothing
        self._enabled: bool = True
        # Hook registry: event name → list of hook components
        self.hooks: Dict[str, List[SpxComponent]] = {}
        self._populate(self.definition)
        self.logger.debug(f"Created {self.__class__.__name__}(name={self.name}, definition={self.definition!r})")

    def add_child(self, child: "SpxComponent"):
        """Prevent adding itself as a child"""
        if child is self:
            raise ValueError(f"Cannot add {self.name} as its own child.")
        if not isinstance(child, SpxComponent):
            raise ValueError("Only SpxComponent instances can be added as children.")
        # add or replace child by its name
        self.children[child.name] = child
        child.parent = self

    def remove_child(self, child: "SpxComponent"):
        if child.name in self.children:
            self.children.pop(child.name)
            child.parent = None

    def get_children(self) -> Dict[str, "SpxComponent"]:
        """Return the internal dict of children keyed by name."""
        return self.children

    def get_children_list(self) -> List["SpxComponent"]:
        """Return a list of child components."""
        return list(self.children.values())

    def get_parent(self):
        return self.parent

    def get_hierarchy(self):
        return {
            "name": self.name,
            "children": [child.get_hierarchy() for child in self.children.values()],
        }

    def get_root(self):
        if self.parent is None:
            return self
        return self.parent.get_root()

    def prepare(self, *args, **kwargs) -> bool:
        """Prepare the component and propagate to children."""
        if not self._enabled:
            # Trigger on_prepare hooks
            self.trigger_hooks("on_event", *args, **kwargs)
            self.trigger_hooks("on_prepare", *args, **kwargs)
            self.logger.debug(f"Component {self.name} is disabled; skipping prepare")
            return False
        # Trigger on_prepare hooks
        self.trigger_hooks("on_event", *args, **kwargs)
        self.trigger_hooks("on_prepare", *args, **kwargs)
        self.logger.debug(f"Preparing {self.name}")
        self.state = SpxComponentState.PREPARING
        for child in self.children.values():
            child.prepare(*args, **kwargs)
        self.state = SpxComponentState.PREPARED
        return True

    def run(self, *args, **kwargs) -> bool:
        """Run the component and propagate to children."""
        if not self._enabled:
            # Trigger on_run hooks
            self.trigger_hooks("on_event", *args, **kwargs)
            self.trigger_hooks("on_run", *args, **kwargs)
            self.logger.debug(f"Component {self.name} is disabled; skipping run")
            return False
        # Trigger on_run hooks
        self.trigger_hooks("on_event", *args, **kwargs)
        self.trigger_hooks("on_run", *args, **kwargs)
        self.logger.debug(f"Running {self.name}")
        self.state = SpxComponentState.RUNNING
        for child in self.children.values():
            child.run(*args, **kwargs)
        self.state = SpxComponentState.STOPPED
        return True

    def start(self, *args, **kwargs) -> bool:
        """Start the component and propagate to children."""
        if not self._enabled:
            self.logger.debug(f"Component {self.name} is disabled; skipping start")
            return False
        # Trigger on_start hooks
        self.trigger_hooks("on_event", *args, **kwargs)
        self.trigger_hooks("on_start", *args, **kwargs)
        self.logger.debug(f"Starting {self.name}")
        self.state = SpxComponentState.STARTING
        for child in self.children.values():
            child.start(*args, **kwargs)
        self.state = SpxComponentState.STARTED
        return True

    def pause(self, *args, **kwargs) -> bool:
        """Pause the component and propagate to children."""
        self.logger.debug(f"Pausing {self.name}")
        self.state = SpxComponentState.PAUSING
        for child in self.children.values():
            child.pause(*args, **kwargs)
        self.state = SpxComponentState.PAUSED
        return True

    def stop(self, *args, **kwargs) -> bool:
        """Stop the component and propagate to children."""
        if not self._enabled:
            self.logger.debug(f"Component {self.name} is disabled; skipping stop")
            return False
        self.logger.debug(f"Stopping {self.name}")
        self.state = SpxComponentState.STOPPING
        for child in self.children.values():
            child.stop(*args, **kwargs)
        self.state = SpxComponentState.STOPPED
        return True

    def reset(self, *args, **kwargs) -> bool:
        """Reset the component and propagate to children."""
        self.logger.debug(f"Reset {self.name}")
        self.state = SpxComponentState.RESETTING
        for child in self.children.values():
            child.reset(*args, **kwargs)
        self.state = SpxComponentState.RESET
        return True

    def destroy(self, *args, **kwargs) -> bool:
        """Destroy the component and propagate to children."""
        self.logger.debug(f"Destroying {self.name}")
        self.state = SpxComponentState.DESTROYING
        for child in self.children.values():
            child.destroy(*args, **kwargs)
        self.parent = None
        self.children = {}
        self.state = SpxComponentState.DESTROYED
        return True

    def __repr__(self):
        return f"{self.__class__.__name__}(name='{self.name}', state={self.state.name}, parent={self.parent.__class__.__name__ if self.parent else None}, children={len(self.children)})"

    def _populate(self, definition: dict):
        """
        Generic hook to populate instance attributes from definition dict.
        By default, this will set each key in definition as an attribute.
        Subclasses can override to customize behavior.
        """
        if isinstance(definition, dict):
            if dataclasses.is_dataclass(self):
                field_names = {field.name for field in dataclasses.fields(self)}
                for key, value in definition.items():
                    if key in field_names:
                        setattr(self, key, value)
                    else:
                        path = self._get_full_path()
                        raise AttributeError(
                            f"Cannot set undefined dataclass field '{key}' on '{path}'"
                        )
            else:
                for key, value in definition.items():
                    if hasattr(self, key):
                        setattr(self, key, value)
                    else:
                        path = self._get_full_path()
                        raise AttributeError(
                            f"Cannot set undefined attribute '{key}' on '{path}'"
                        )

    def _get_full_path(self) -> str:
        """
        Construct the full component path by joining all parent names with dots.
        Returns a string like 'root.child.grandchild'.
        """
        path = self.name
        node = self.parent
        while node:
            path = f"{node.name}.{path}"
            node = node.parent
        return path

    def __getitem__(self, key: str):
        """
        If this component has children, retrieve a child by name.
        If no children exist, treat the key as an attribute name on this instance.
        Raises KeyError if neither a child nor an attribute with that name exists.
        """
        if len(self.children) > 0:
            # Normal child lookup
            try:
                return self.children[key]
            except KeyError:
                raise KeyError(f"No child named '{key}' in component '{self.name}'.")
        else:
            # Leaf: try to return attribute
            if hasattr(self, key):
                return getattr(self, key)
            raise KeyError(f"Component '{self.name}' has no child or attribute named '{key}'.")

    def add(self, inst_name: str, cfg: Any) -> "SpxComponent":
        """
        Dynamically add a new instance at runtime, mirroring _populate logic for a single entry.
        Returns the newly created component (or existing if already present).
        """
        # If already present, remove existing instance first
        self.remove(inst_name)
        # Reuse _populate logic by feeding a single-entry list
        self._populate([{inst_name: cfg}])
        # Return the newly added instance
        return self.children.get(inst_name)

    def __setitem__(self, key: str, value: Any) -> None:
        if isinstance(value, SpxComponent):
            # Assign a child component
            old = self.children.get(key)
            if old is not None:
                old.parent = None
            self.children[key] = value
            value.parent = self
            return
        # Override existing attribute if present
        if hasattr(self, key):
            cls_attr = getattr(self.__class__, key, None)
            if isinstance(cls_attr, property):
                # Property: ensure it has a setter
                if cls_attr.fset is None:
                    raise AttributeError(f"Property '{key}' on component '{self.name}' is read-only")
                setattr(self, key, value)
            else:
                setattr(self, key, value)
            return
        # For dict or string values, delegate to add()
        if isinstance(value, (dict, str)):
            self.add(key, value)
            return
        # Otherwise, cannot handle this type
        raise ValueError(f"Cannot set item {key!r} with value of type {type(value).__name__}")

    def remove(self, inst_name: str) -> None:
        """
        Remove the instance with the given name, if present.
        """
        child = self.children.pop(inst_name, None)
        if child is not None:
            # detach parent
            child.parent = None

    def __delitem__(self, key: str) -> None:
        """
        Allow dict-style deletion of instances.
        """
        self.remove(key)

    def get(self, key: str, default=None) -> Optional["SpxComponent"]:
        """
        Get a child component by name, returning a default if not found.

        Args:
            key (str): Name of the child component.
            default (Any): Value to return if the child is not found.

        Returns:
            Optional[SpxComponent]: The child component or the default.
        """
        return self.children.get(key, default)

    def __contains__(self, key: str) -> bool:
        """
        Check if a child component exists by name.

        Args:
            key (str): Name to check.

        Returns:
            bool: True if a child with the given name exists, False otherwise.
        """
        return key in self.children

    def __len__(self) -> int:
        """
        Return the number of child components.

        This allows using len(component) to retrieve the count of children.

        Returns:
            int: Number of child components.
        """
        return len(self.children)

    def __bool__(self) -> bool:
        """
        Always return True to ensure component truthiness is independent of child count.
        """
        return True

    def enable(self) -> None:
        """
        Enable the component. Lifecycle methods (prepare, run, start, stop)
        will execute when this component is enabled.
        """
        self._enabled = True
        # Trigger enable hooks
        self.trigger_hooks("on_event")
        self.trigger_hooks("on_enable")

    def disable(self) -> None:
        """
        Disable the component. Lifecycle methods (prepare, run, start, stop)
        become no-ops when the component is disabled.
        """
        self._enabled = False
        # Trigger disable hooks
        self.trigger_hooks("on_event")
        self.trigger_hooks("on_disable")

    @property
    def enabled(self) -> bool:
        """
        Whether the component is enabled. If False, prepare/run/start/stop
        will not execute.
        """
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool) -> None:
        """
        Setter for enabled flag. Allows toggling via assignment:
            component.enabled = False  # equivalent to component.disable()
        """
        if value:
            self.enable()
        else:
            self.disable()

    def register_hook(self, event: str, hook_component: "SpxComponent") -> None:
        """
        Register a hook component to be triggered on the given event.
        Prevents duplicate registrations of the same hook for the same event.
        """
        hooks_list = self.hooks.setdefault(event, [])
        if hook_component not in hooks_list:
            hooks_list.append(hook_component)

    def get_hooks(self, event: str) -> List["SpxComponent"]:
        """
        Get all registered hook components for the given event name.
        """
        return list(self.hooks.get(event, []))

    def trigger_hooks(self, event: str, *args, **kwargs) -> None:
        """
        Invoke all hooks registered under the given event.
        """
        for hook in self.hooks.get(event, []):
            try:
                hook.run(*args, **kwargs)
            except Exception as e:
                self.logger.error(f"Hook '{hook.name}' on event '{event}' failed: {e}")



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from typing import Any, Dict
from spx_sdk.components import SpxComponent, SpxComponentState
from spx_sdk.attributes.resolve_attribute import (
    substitute_attribute_references_hierarchical,
    resolve_attribute_reference_hierarchical,
    ATTRIBUTE_REFERENCE_PATTERN
)
from spx_sdk.attributes.resolve_attribute import (
    extract_attribute_wrappers_hierarchical,
    substitute_with_wrappers
)


class Action(SpxComponent):
    """
    Base class for a single action mapping inputs to a function call and writing an output.
    Definition keys:
      - function (str): name of the function to call
      - output (str): hash-ref string for output attribute
      - param_1 (str, Any): param_1 name -> hash-ref string or value for input
      - param_2 (str, Any): param_2 name -> hash-ref string or value for input
      - ...
    """
    def _populate(self, definition: Dict) -> None:
        self.function: str = None
        self.output: Any = None  # Can be a single string or a list of strings
        self.outputs: dict = {}  # Will be populated with attribute instances keyed by name
        # 1) Extract any extra keys (besides 'function' and 'output') into self.params
        #    so that subclasses can use them later.
        self.params = {}
        for key in list(definition.keys()):
            if key not in ("function", "output"):
                self.params[key] = definition.pop(key)

        # 1.5) Prepare empty wrapper mapping: key -> list of (ref, wrapper)
        self._wrappers: Dict[str, list] = {}

        # 2) Delegate to parent to populate "function" and "output" attributes, etc.
        super()._populate(definition)

        # Resolve outputs into a dict of attribute instances keyed by attribute name, if any
        if self.output:
            refs = self.output if isinstance(self.output, list) else [self.output]
            for ref in refs:
                if ref is None:
                    continue
                attr = resolve_attribute_reference_hierarchical(self.parent, ref)
                if attr is None:
                    raise ValueError(f"Output reference '{ref}' could not be resolved.")
                # Use the attribute's name as the dictionary key
                self.outputs[attr.name] = attr
        self.state = SpxComponentState.INITIALIZED

    def resolve_param(self, param):
        if not isinstance(param, str):
            return param
        # Substitute attribute references (e.g. "$attr(x)") with literal values
        expr = substitute_attribute_references_hierarchical(self.parent, param)
        # Normalize boolean literals
        expr = expr.replace("true", "True").replace("false", "False")
        # Evaluate the resulting expression to compute numeric or boolean results
        try:
            return eval(expr)
        except Exception:
            return expr

    def prepare(self, *args, **kwargs) -> bool:
        """
        Defer parameter resolution: collect wrappers for each raw param now.
        """
        if self._enabled is False:
            return True
        self.state = SpxComponentState.PREPARING
        # For every key in self.params, extract wrappers without evaluating
        for key, raw_val in self.params.items():
            # If raw_val is a string containing reference syntax, extract wrappers
            if isinstance(raw_val, str) and ATTRIBUTE_REFERENCE_PATTERN.search(raw_val):
                wrappers = extract_attribute_wrappers_hierarchical(self.parent, raw_val)
            else:
                # static literal or string without references: assign immediately, cast if possible
                if isinstance(raw_val, str):
                    try:
                        evaluated = eval(raw_val)
                    except Exception:
                        evaluated = raw_val
                else:
                    evaluated = raw_val
                setattr(self, key, evaluated)
                wrappers = []
            self._wrappers[key] = wrappers
        self.state = SpxComponentState.PREPARED
        return True

    def apply_wrappers(self):
        """
        Resolve raw param expressions by substituting attribute values.
        Sets self.<key> to the resolved value (eval if possible).
        """
        for key, raw_val in self.params.items():
            wrappers = self._wrappers.get(key, [])
            # Only reapply for dynamic parameters (with wrappers)
            if not wrappers:
                continue
            # Substitute wrapper values into the raw text
            substituted = substitute_with_wrappers(raw_val, wrappers)
            # Try to evaluate numeric/boolean expressions, fallback to string
            try:
                resolved = eval(substituted)
            except Exception:
                resolved = substituted
            setattr(self, key, resolved)

    def write_outputs(self, result) -> None:
        """
        Write the result value to all output attributes.
        """
        for output in self.outputs.values():
            output.set(result)
        return result

    def run(self, result=None) -> Any:
        if self._enabled is False:
            return True
        self.state = SpxComponentState.RUNNING
        # First, resolve parameters now that wrappers are ready
        self.apply_wrappers()
        if result is None:
            return None  # No result to set, nothing to run

        self.write_outputs(result)
        self.state = SpxComponentState.STOPPED
        return result



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from typing import Any
from spx_sdk.registry import register_class
from spx_sdk.attributes import substitute_attribute_references_hierarchical
from spx_sdk.components import SpxContainer, SpxComponentState


@register_class(name="conditions")
@register_class(name="cases")
@register_class()
class Conditions(SpxContainer):
    """
    SPX container for a list of conditions. Definition must be a list of dicts:
      - <condition_name>: "<condition_ref>"
        <action_1>: ...
        <action_2>: ...
        <actions>: ...
        ...
    Each dict is turned into a Condition child component.
    """
    pass


@register_class(name="if_chain")
class IfChain(SpxContainer):
    """
    A container for a chain of conditions. It runs the first condition that evaluates to true.
    Definition must be a list of dicts, each representing a condition.
    Example:
    [
        {"if": "condition1", "then": "action1"},
        {"if": "condition2", "then": "action2"},
        {"else": "default_action"}
    ]
    Each dict is turned into a Condition child component.
    """
    def _populate(self, definition: Any) -> None:
        super()._populate(definition)

    def run(self, *args, **kwargs) -> bool:
        self.state = SpxComponentState.RUNNING
        context = kwargs.get("context", {})
        for child in self.children.values():
            if child.run(context=context):
                break
        self.state = SpxComponentState.STOPPED
        return True

    def prepare(self, *args, **kwargs) -> bool:
        self.state = SpxComponentState.PREPARING
        context = kwargs.get("context", {})
        for child in self.children.values():
            if child.prepare(context=context):
                break
        self.state = SpxComponentState.PREPARED
        return True


@register_class(name="if")
@register_class(name="when")
@register_class(name="case")
@register_class(name="ifelse")
@register_class(name="if_else")
@register_class(name="else_if")
@register_class(name="elseif")
@register_class(name="condition")
@register_class()
class Condition(SpxContainer):

    def evaluate(self, condition: str) -> bool:
        # Replace #attr, #internal, and #external references with their actual values
        resolved_condition = substitute_attribute_references_hierarchical(self.parent, condition)
        resolved_condition = resolved_condition.replace("true", "True").replace("false", "False")

        try:
            return eval(resolved_condition)
        except Exception as e:
            self.logger.error(f"Error evaluating condition: {resolved_condition}")
            self.logger.exception(e)
            return False

    """
    A condition that runs its children if the condition is met.
    """
    def run(self, *args, **kwargs) -> bool:
        if not self.evaluate(self.definition):
            return False
        return super().run(*args, **kwargs)

    def prepare(self, *args, **kwargs) -> bool:
        if not self.evaluate(self.definition):
            return False
        return super().prepare(*args, **kwargs)


@register_class(name="else")
@register_class()
class Else(SpxContainer):
    def run(self, *args, **kwargs) -> bool:
        return super().run(*args, **kwargs)



# SPDX-License-Identifier: MIT
# Copyright (c) 2025 Hammerheads Engineers Sp. z o.o.
# See the accompanying LICENSE file for terms.

from typing import Any, Dict, List
from spx_sdk.registry import register_class, get_classes_by_base
from spx_sdk.components import SpxContainer
from spx_sdk.actions.action import Action


@register_class(name="actions")
@register_class(name="Actions")
class Actions(SpxContainer):
    """
    SPX container for a list of actions. Definition must be a list of dicts:
      - <function_name>: "<output_ref>"
        <param1>: "<input_ref1>"
        ...
    Each dict is turned into an Action child component.
    """
    def _populate(self, definition: List[Dict[str, Any]]) -> None:
        # Validate definition format
        if not isinstance(definition, list):
            raise ValueError("Actions definition must be a list of mappings")

        # Track occurrences of each action name for unique naming
        name_counts: Dict[str, int] = {}
        # Pre-fetch all subclasses of Action once
        subclasses = get_classes_by_base(Action.__name__)

        for entry in definition:
            if not isinstance(entry, dict) or not entry:
                continue
            # Get the first (function -> output) pair
            fn_name, output_ref = next(iter(entry.items()))
            # Determine a unique child name
            count = name_counts.get(fn_name, 0)
            child_name = fn_name if count == 0 else f"{fn_name}_{count}"
            name_counts[fn_name] = count + 1
            # Collect remaining key/value pairs as params
            params = {k: v for k, v in entry.items() if k != fn_name}
            action_def = {"function": fn_name, "output": output_ref, **params}
            # Instantiate either the specific subclass or default Action
            action_cls = subclasses.get(fn_name)
            if action_cls:
                action_cls(name=child_name, parent=self, definition=action_def)
            else:
                super()._populate(entry)
